/*
  LF_TEST V1.0
  -- Low Frequency Sine-Burst Test Signals for Room and Loudspeaker Evaluation --
  Copyright (C) 2012  Klaus Strohhaecker

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>


typedef unsigned char	byte;      // make sure this is 8-bit wide !
typedef unsigned int	word;      // make sure this is 16-bit wide !
typedef unsigned long	dword;     // make sure this is 32-bit wide !


struct wav_hdr { // 44.100Hz, 16bit stereo WAV header
	char	h1,h2,h3,h4;		//"RIFF"
	dword	tsize_m8;       	// file size -8
	char	h5,h6,h7,h8;		// "WAVE"
	char	h9,h10,h11,h12;		// "fmt_"
	dword	fmt_size;		// 0x10ul
	word	c1;			// 0x0001
	word	channels;		// 2
	dword	sample_rate;		// 44100
	dword	bytes_per_sec;		// 176400
	word	bytes_per_sample;	// 4
	word	bits_per_sample;	// 16
	char	h13,h14,h15,h16;	// "data"
	dword	tsize_m44;		// file size -44
};

const double 	srate=44100.0;

struct wav_hdr header = { // filled WAV-header
	'R','I','F','F',
	0,
	'W','A','V','E',
	'f','m','t',' ',
	16ul,
	1u,
	2,
	44100ul,
	176400ul,
	4,
	16,
	'd','a','t','a',
	0
};
const	HSIZE=sizeof(header);





FILE	*fp;
char	_fname[300];

void wopen(const char *fname)
{
	strcpy(_fname, fname);		// _fname := name
	fp=fopen(fname, "w+b");
	if (fp==NULL)
		exit(3);
	if (fwrite(&header, HSIZE, 1, fp) != 1)
		exit(3);
}

void wclose(void)
{
	fflush(fp);
	dword len=ftell(fp);
	if (fclose(fp) == EOF)
		exit(3);
	fp=fopen(_fname, "r+b");
	if (fp==NULL)
		exit(3);
	header.tsize_m8 = len-8;
	header.tsize_m44 = (len-HSIZE);
	if (fwrite(&header, HSIZE, 1, fp) != 1)
		exit(3);
	if (fclose(fp) == EOF)
		exit(3);
}


inline void wr(int sample)
{
	fwrite(&sample, 2, 1, fp);
	fwrite(&sample, 2, 1, fp);
}

inline void wr2(int sampleL, int sampleR)
{
	fwrite(&sampleL, 2, 1, fp);
	fwrite(&sampleR, 2, 1, fp);
}






inline double XexpY(double x, double y)
{
	return exp(log(x) * y);
}


void burst(
	double		freq		    // base frequency
       ,dword		cycles=20	    // # of cycles per burst
       ,dword		bursts=12	    // # of bursts
       ,dword		envel=2		    // # of envelopped cycles thereof
       ,double		spacing_phi=0.0	    // phase offset between bursts
       ,double		L_mul = 1.0	    // left channel multiplier
       ,double		R_mul = 1.0	    // right channel multiplier
       ,double		cycle_decay=1.0   // decay rate of cycle count
       ,double		spacing=0.3	    // spacing between bursts
       ,double		spacing_decay=0.6  // decay rate thereof
       ,double		lead_in=0.1	    // lead in time (secs)
       ,double		lead_out=1.0	    // lead out time (secs)
) {
	if (freq > 22000.0)	freq = 22000.0;
	if (freq < 10.0)	freq = 10.0;
	if (bursts < 1)		bursts = 1;
	if (cycles < 6)		cycles = 6;
	if (envel > cycles/3)	envel = cycles/3;
	if (envel < 1)		envel = 1;
	if (cycle_decay <= 0.1)	cycle_decay = 0.1;
	if (cycle_decay > 1.0)	cycle_decay = 1.0;
	spacing=cycles/freq;
	if (spacing < 0.002)	spacing = 0.002;
	if (spacing > 1.0)	spacing = 1.0;
	if (spacing_decay <0.1)	spacing_decay = 0.1;
	if (spacing_decay >1.0)	spacing_decay = 1.0;
	if (lead_in < 0.01)	lead_in = 0.01;
	if (lead_in > 2.0)	lead_in = 2.0;
	if (lead_out < 0.1)	lead_out = 0.1;
	if (lead_out > 5.0)	lead_out = 5.0;
	if (spacing_phi < 0.0)	spacing_phi = 0.0;
	if (spacing_phi > 1.0)	spacing_phi = 1.0;

	long    k;
	double	s_f	= srate/freq;

// lead in silence
	for (k=0;  k < lead_in*srate;  k++)
		wr(0);

//.....
	for (dword b=0;  b<bursts;  b++)
	{ // all bursts
		if (b>0)
		{ // all but first burst
			if (spacing < 0.002)
				spacing = 0.002;
			k = long(0.5 + spacing*freq);
			if (k==0)
				k++,k--;
			k = long((k+spacing_phi) / freq * srate);
			for (;  k>0;  k--)
				wr(0); // inter-burst silence
			spacing = spacing * spacing_decay;
			cycles = cycles * cycle_decay;
			if (cycles < 6)		cycles = 6;
			if (envel > cycles/3)	envel = cycles/3;
		}
		for (k=0;  k < s_f*cycles; k++)
		{ // all samples in all cycles
			double m=1.0;
			double x;
			int c = k / s_f;
			if (c < envel)
			{ // fade in
				x = k / (envel*s_f);
				m = -cos(M_PI*x)/2.0 + 0.5;
			}
			else if (c >= cycles-envel)
			{ // fade out
				x = k - (cycles-envel)*s_f;
				x /= (envel*s_f);
				m = -cos(M_PI*(1-x))/2.0 + 0.5;
			}

			x = m*32760.0*sin(2.0*M_PI*k/s_f);
			wr2(x*L_mul, x*R_mul);
		} // samples

	} // bursts


// lead out silence
	for (k=0;  k < lead_out*srate;  k++)
		wr(0);
}





int main(void)
{
	      double	fmin = 27.5;		       	// VERY LOW A note
	const double 	fmax = fmin * 2 * 2 * 2 * 2;	// 4 oct (440Hz)
	const double	fratio = fmax/fmin;
	      int	count = 4 * 12 * 2;             // 4 oct * 24/oct
	const double	m = XexpY(fratio, 1.0/count);

	double	fact=1.0;

	char fname[]="12345p7Hz.WAV";

// obs.	for (int k=0; k<=OCTAVES*TONES_PER_OCT; k++)
//	count=0; fmin=220;
	for (int k=0; k<= count; k++)
	{

		double freq=fmin*fact;
		if (freq > 22000.0)
			break;
		fact*=m;
		sprintf(fname, "%07.1lfH.WAV", freq);
		fname[5]='p';



#if 1
// create file
		printf("writing %s ...", fname);
		wopen(fname);

		burst(
			freq
		       ,freq*0.4		// first burst 1.0sec
		       ,20			// 12 bursts
		       ,2			// 2 envelopped cycles
		       ,0.0			// 0 degree spacing phase
		     );
		burst(
			freq
		       ,freq*0.4		// first burst 1.0sec
		       ,20			// 12 bursts
		       ,2			// 2 envelopped cycles
		       ,0.5			// 180 degree spacing phase
		     );
		burst(
			freq
		       ,freq*0.4		// first burst 1.0sec
		       ,20			// 12 bursts
		       ,2			// 2 envelopped cycles
		       ,0.0			// 0 degree spacing phase
		       ,1.0			// Left channel normal
		       ,-1.0			// Right channel inverted
		     );
		burst(
			freq
		       ,freq*0.4		// first burst 1.0sec
		       ,20			// 12 bursts
		       ,2			// 2 envelopped cycles
		       ,0.5			// 180 degree spacing phase
		       ,1.0			// Left channel normal
		       ,-1.0			// Right channel inverted
		     );

// flush file and update header length fields
		wclose();

		printf("\n");
#else
// create index
		printf("%2d; %s\n", k+1, fname);
#endif
	}

	printf("\nsuccess!\n");
	return 0;
}
