/*
  LF_TEST V1.0
  -- Low Frequency Sine-Burst Test Signals for Room and Loudspeaker Evaluation --
  Copyright (C) 2012  Klaus Strohhaecker

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>


typedef unsigned char	byte;      // make sure this is 8-bit wide !
typedef unsigned int	word;      // make sure this is 16-bit wide !
typedef unsigned long	dword;     // make sure this is 32-bit wide !


struct wav_hdr { // 44.100Hz, 16bit stereo WAV header
	char	h1,h2,h3,h4;		//"RIFF"
	dword	tsize_m8;       	// file size -8
	char	h5,h6,h7,h8;		// "WAVE"
	char	h9,h10,h11,h12;		// "fmt_"
	dword	fmt_size;		// 0x10ul
	word	c1;			// 0x0001
	word	channels;		// 2
	dword	sample_rate;		// 44100
	dword	bytes_per_sec;		// 176400
	word	bytes_per_sample;	// 4
	word	bits_per_sample;	// 16
	char	h13,h14,h15,h16;	// "data"
	dword	tsize_m44;		// file size -44
};

const double 	srate=44100.0;
const long	israte=long(srate);

struct wav_hdr header = { // filled WAV-header
	'R','I','F','F',
	0,
	'W','A','V','E',
	'f','m','t',' ',
	16ul,
	1u,
	2,
	44100ul,
	176400ul,
	4,
	16,
	'd','a','t','a',
	0
};
const	HSIZE=sizeof(header);





FILE	*fp;
char	_fname[300];

void wopen(const char *fname)
{
	strcpy(_fname, fname);		// _fname := name
	fp=fopen(fname, "w+b");
	if (fp==NULL)
		exit(3);
	if (fwrite(&header, HSIZE, 1, fp) != 1)
		exit(3);
}

void wclose(void)
{
	fflush(fp);
	dword len=ftell(fp);
	if (fclose(fp) == EOF)
		exit(3);
	fp=fopen(_fname, "r+b");
	if (fp==NULL)
		exit(3);
	header.tsize_m8 = len-8;
	header.tsize_m44 = (len-HSIZE);
	if (fwrite(&header, HSIZE, 1, fp) != 1)
		exit(3);
	if (fclose(fp) == EOF)
		exit(3);
}


inline void wr(int sample)
{
	fwrite(&sample, 2, 1, fp);
	fwrite(&sample, 2, 1, fp);
}






inline double XexpY(double x, double y)
{
	return exp(log(x) * y);
}


void sweep(
	const char	*fname		    // file name
       ,double		fstart=100	    // start frequency
       ,double		fend=4000	    // end frequency
       ,double		time=10.0	    // sweep time
       ,double		lead_in=0.2	    // lead in time (secs)
       ,double		lead_out=0.2	    // lead out time (secs)
) {
	if (fstart > 22000.0)	fstart = 22000.0;
	if (fstart < 10.0)	fstart = 10.0;
	if (fend > 22000.0)	fend = 22000.0;
	if (fend < 10.0)	fend = 10.0;
	if (time < 1.0)		time = 1.0;
	if (time > 600.0)	time = 600.0;
	if (lead_in < 0.1)	lead_in = 0.1;
	if (lead_in > 5.0)	lead_in = 5.0;
	if (lead_out < 0.1)	lead_out = 0.1;
	if (lead_out > 5.0)	lead_out = 5.0;

	dword k;
	double x, f, m;

	double samples	= time*srate;
	dword  isamples = samples;
	double fratio=fend/fstart;


// create file
	wopen(fname);

// lead in silence
	long cntr=lead_in*srate;
	int  sect=0;
	for (k=0;  k < cntr;  k++)
		wr(0);
	while (cntr >= israte)
	{
		cntr -= israte;
		sect++;
	}

//.....
//	double dummy;		// needed for modf()
	x = 0.0;		// angle x runs through [0, 1[
	for (dword s=0; s<isamples;  s++)
	{ // all samples
		// determine frequency of next sample
		f = fstart * XexpY(fratio, s/samples);
		// next angle to derive a sample from
		x += f/srate; // this of course accumulates some error

		if (cntr==0)
			printf("%3d; %5.1lf Hz\n", sect++, f);
		if (++cntr >= srate)
			cntr=0;
#if 1
		// create cosine envelope for low distortion fade in/out
		m = 1.0;
		if (s < 4410) // 0.1 sec fade in
			m = -cos(M_PI * s/4410.0)/2+.5;
		else if (isamples-s < 4410) // 0.1 sec fade out
			m = -cos(M_PI * (isamples-s)/4410.0)/2+.5;

		m *= XexpY(0.25, s/samples);
		wr(m * 32760.0*sin(2.0*M_PI* x));
//		x = modf(x, &dummy); // keep fractional part of x;
#endif
	} // samples


// lead out silence
	for (k=0;  k < lead_out*srate;  k++)
		wr(0);

// flush file and update header length fields
	wclose();
}





int main(void)
{
	const double	fmin = 27.5;
	const double 	fmax = 440;
	const double	time = 96;
	char fname[]=	"TEST.WAV";

	printf("writing %s ...\n", fname);

	sweep(fname, fmin, fmax, time, 3);

	printf("success!\n");
	return 0;
}
